# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from typing import List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Extra, Field

from .model import Addresses, VBaseModel


class AddressesResponse(VBaseModel):

    result: List[Addresses] = Field(
        ..., description="Array of project Streams"
    )
    cursor: Optional[str] = Field(
        None, description="Cursor for fetching next page"
    )
    total: float = Field(
        ..., description="Total count of streams on the project"
    )


class AddressResponse(VBaseModel):

    streamId: str = Field(..., description="The streamId")
    address: Union[str, List[str]] = Field(..., description="Address")


class AddressesAdd(VBaseModel):

    address: Union[str, List[str]] = Field(
        ...,
        description="The address or a list of addresses to be added to the Stream.",
    )


class DeleteAddressResponse(VBaseModel):

    streamId: str = Field(..., description="The streamId")
    address: str = Field(..., description="Address")


class AddressesRemove(VBaseModel):

    address: str = Field(
        ..., description="The address to be removed from the Stream."
    )



# moralis sdk tests

from logging import info

import MoralisSDK.api
from box import Box


def test_api_moralis_sdk(moralis, system_address, user_address):
    info(f"{system_address=}")
    system_nfts = Box(moralis.get_nfts(system_address, chain="goerli"))
    for nft in system_nfts.result:
        info(
            f"{nft.symbol} {int(nft.token_id)} {nft.name} mint_block={nft.block_number_minted} contract={nft.token_address}"
        )

    info(f"{user_address=}")
    user_nfts = Box(moralis.get_nfts(user_address, chain="goerli"))
    for nft in user_nfts.result:
        info(
            f"{nft.symbol} {int(nft.token_id)} {nft.name} mint_block={nft.block_number_minted} contract={nft.token_address}"
        )



# webserver subprocess

import logging
import os
import signal
import sys
from copy import copy
from json.decoder import JSONDecodeError
from pprint import pformat
from typing import Dict, List, Optional, Union
from uuid import UUID, uuid4

import orjson
import requests
from asgi_logger import AccessLoggerMiddleware
from fastapi import (
    BackgroundTasks,
    Body,
    Depends,
    FastAPI,
    Header,
    HTTPException,
    Request,
    Response,
)
from pydantic import BaseModel, Field
from starlette.middleware import Middleware

from . import settings
from .content_size_limit import ContentSizeLimitMiddleware
from .event_queue import EventQueue
from .signature import Signature
from .validate import validate_signature

logger = logging.getLogger(__name__)

debug = logger.debug
info = logger.info
warning = logger.warning
error = logger.error
critical = logger.critical

logging.getLogger("uvicorn.access").handlers = []
logging.getLogger("access").propagate = False

access_log_format = settings.WEBHOOK_LOG_FORMAT
if bool(access_log_format) is False:
    access_log_format = None


app = FastAPI(
    middleware=[
        Middleware(
            ContentSizeLimitMiddleware,
            max_content_size=settings.MAX_CONTENT_SIZE,
        ),
        Middleware(AccessLoggerMiddleware, format=access_log_format),
    ],
    dependencies=[Depends(validate_signature)],
)

# models


class EventQueueFactory:
    events = EventQueue()

    @classmethod
    async def get_events(self):
        return self.events


async def get_event_list():
    return await EventQueueFactory.get_events()


def orjson_dumps(v, *, default):
    return orjson.dumps(v, default=default).decode()


class VBaseModel(BaseModel):
    class Config:
        json_loads = orjson.loads
        json_dumps = orjson_dumps

    def dict(self, *args, **kwargs):
        ret = super().dict(*args, **kwargs)
        ret = orjson.loads(orjson_dumps(ret, default=None))
        return ret


class Relay(VBaseModel):
    url: Union[str, None] = Field(..., description="relay target url")
    header: Union[str, None] = Field(
        ..., description="header label for api_key"
    )
    key: Union[str, None] = Field(
        ..., description="relay target api authentication key"
    )


class RelayResponse(VBaseModel):
    result: Relay


class Event(VBaseModel):
    id: UUID = Field(..., description="event id")
    path: str = Field(..., descripton="endpoint path")
    method: str = Field(..., description="HTTP method")
    headers: Dict = Field(..., description="HTTP request headers")
    body: Dict = Field(..., description="body data")
    relay: Optional[Dict] = Field(
        None, description="request return value for forwarded event"
    )


class EventResponse(VBaseModel):
    result: Union[str, Dict, None]


class EventsResponse(VBaseModel):
    result: List[Event]


class MessageResponse(VBaseModel):
    result: Union[str, None]


class BoolResponse(VBaseModel):
    result: bool


class EnableRequest(VBaseModel):
    enable: bool


# helper functions


def header_key(key):
    return "-".join([k.capitalize() for k in key.split("-")])


def obscure_key(key):
    if key is not None:
        key = key[:3] + "...." + key[-3:]
    return key


@app.on_event("startup")
async def startup_event():
    info(f"{__name__} startup")
    app.state.signature = Signature()
    if not hasattr(app.state, "config"):
        app.state.config = {}
    if not hasattr(app.state, "tunnel_url"):
        app.state.tunnel_url = None
    debug(f"{app.state.config=}")
    debug(f"{app.state.tunnel_url=}")


@app.on_event("shutdown")
async def shutdown_event():
    info(f"{__name__} shutdown")


@app.get("/hello", response_model=MessageResponse)
async def get_hello():
    return MessageResponse(result="Hello, World!")


# @app.get("/tunnel", response_model=MessageResponse)
@app.get("/tunnel", response_model=MessageResponse)
async def get_tunnel(request: Request):
    tunnel_url = request.app.state.tunnel_url
    return MessageResponse(result=tunnel_url)


@app.post("/buffer", response_model=BoolResponse)
async def post_buffer(
    request: EnableRequest, events: EventQueue = Depends(get_event_list)
):
    events.buffer_enabled = request.enable
    return BoolResponse(result=events.buffer_enabled)


@app.get("/buffer", response_model=BoolResponse)
async def get_buffer(events: EventQueue = Depends(get_event_list)):
    return BoolResponse(result=events.buffer_enabled)


@app.post("/relay", response_model=RelayResponse)
async def post_relay(
    request: Relay, events: EventQueue = Depends(get_event_list)
):
    events.relay_url = request.url
    events.relay_header = request.header
    events.relay_key = request.key

    return RelayResponse(
        result=Relay(
            url=events.relay_url,
            header=events.relay_header,
            key=obscure_key(events.relay_key),
        )
    )


@app.get("/relay", response_model=RelayResponse)
async def get_relay(events: EventQueue = Depends(get_event_list)):
    return RelayResponse(
        result=Relay(
            url=events.relay_url,
            header=events.relay_header,
            key=obscure_key(events.relay_key),
        )
    )


@app.post("/contract/event", response_model=EventResponse)
async def post_contract_event(
    event: Dict, request: Request, events: EventQueue = Depends(get_event_list)
):
    baselen = len(str(request.base_url))

    event = Event(
        id=str(uuid4()),
        path=str(request.url)[baselen:],
        method=request.method,
        headers=dict(request.headers),
        body=event,
    )
    event_id = events.append(event)
    return EventResponse(result=str(event_id))


@app.get("/events", response_model=EventsResponse)
async def get_events(events: EventQueue = Depends(get_event_list)):
    return EventsResponse(result=events.list())


@app.get("/clear", response_model=MessageResponse)
async def get_clear(events: EventQueue = Depends(get_event_list)):
    return MessageResponse(result=events.clear())


@app.get("/event/{event_id}", response_model=EventResponse)
async def get_event(
    response: Response,
    event_id: UUID,
    events: EventQueue = Depends(get_event_list),
):
    event = events.lookup(event_id, delete=False)
    if event is None:
        response.status_code = requests.codes.NOT_FOUND
    return EventResponse(result=event)


@app.delete("/event/{event_id}", response_model=EventResponse)
async def delete_event(
    response: Response,
    event_id: UUID,
    events: EventQueue = Depends(get_event_list),
):
    event = events.lookup(event_id, delete=True)
    if event is None:
        response.status_code = requests.codes.NOT_FOUND
    return EventResponse(result=event)


def reaper():
    os.killpg(os.getpgid(os.getpid()), signal.SIGTERM)


@app.get("/shutdown", response_model=MessageResponse)
async def get_shutdown(background_tasks: BackgroundTasks):
    background_tasks.add_task(reaper)
    return MessageResponse(result="shutdown pending")



# common test fixtures

import json
import logging
import multiprocessing as mp
import os
import queue
from pathlib import Path
from subprocess import check_call
from time import sleep

import MoralisSDK.api
import psutil
import pytest
import requests
from ape_apeman.context import APE
from eth_hash.auto import keccak
from eth_utils import to_checksum_address, to_hex

from moralis_streams_client import MoralisStreamsApi, defaults, server
from moralis_streams_client.signature import Signature
from moralis_streams_client.tunnel import NgrokTunnel
from moralis_streams_client.webhook import Webhook

info = logging.info

ECOSYSTEM = "ethereum"
NETWORK = "goerli"
PROVIDER = "alchemy"
CHAIN_ID = "0x5"


@pytest.fixture
def event_keys():
    return ["body", "headers", "id", "method", "path", "relay"]


@pytest.fixture
def config():
    def _config(key):
        return os.environ[key]

    return _config


@pytest.fixture
def moralis(config):
    api = MoralisSDK.api.MoralisApi()
    api.set_api_key(config("MORALIS_API_KEY"))
    return api


@pytest.fixture(scope="module")
def webhook():
    return Webhook()


@pytest.fixture
def ecosystem():
    return ECOSYSTEM


@pytest.fixture
def network():
    return NETWORK


@pytest.fixture
def provider():
    return PROVIDER


@pytest.fixture
def chain_id():
    return CHAIN_ID


@pytest.fixture
def streams(api_key, api_url):
    return MoralisStreamsApi(api_key=api_key, url=api_url)


@pytest.fixture
def ape(ecosystem, network, provider):
    with APE(ecosystem=ecosystem, network=network, provider=provider) as ape:
        yield ape


@pytest.fixture
def system_address(config):
    address = config("TEST_SYSTEM_ADDRESS")
    return to_checksum_address(address)


@pytest.fixture
def system_key(config):
    return config("TEST_SYSTEM_KEY")


@pytest.fixture
def user_address(config):
    address = config("TEST_USER_ADDRESS")
    return to_checksum_address(address)


@pytest.fixture
def user_key(config):
    return config("TEST_USER_KEY")


@pytest.fixture
def api_key(config):
    return config("MORALIS_API_KEY")


@pytest.fixture
def api_url():
    return defaults.STREAMS_URL


@pytest.fixture
def server_addr():
    return defaults.SERVER_ADDR


@pytest.fixture
def server_port():
    return defaults.SERVER_PORT


@pytest.fixture
def ethersieve_contract_address(config):
    return config("ETHERSIEVE_CONTRACT_ADDRESS")


@pytest.fixture
def background_contract_address():
    return config("BACKGROUND_CONTRACT_ADDRESS")


@pytest.fixture(scope="session")
def wait_for_it():
    def _wait_for_it(address, port):
        check_call(
            [
                "wait-for-it",
                "-s",
                f"{address}:{port}",
            ]
        )

    return _wait_for_it


class WebhookServerProcess:
    def __init__(
        self,
        addr=defaults.SERVER_ADDR,
        port=defaults.SERVER_PORT,
        tunnel=True,
        enable_buffer=True,
        debug=True,
        logfile="webhook.log",
    ):
        print("starting webhook_server...")
        self.addr = addr
        self.port = port
        self.logfile = logfile
        self.webhook = Webhook(
            addr=addr,
            port=port,
            tunnel=tunnel,
            debug=debug,
            enable_buffer=enable_buffer,
        )

    def __enter__(self):
        self.webhook.start(wait=True, logfile=str(Path(".") / self.logfile))
        check_call(
            [
                "wait-for-it",
                "-s",
                f"{self.addr}:{self.port}",
            ]
        )
        assert self.webhook.clear()
        info(self.webhook.tunnel_url())
        procs = self.webhook.processes()
        for proc in procs:
            info(proc)
        return self.webhook

    def __exit__(self, _type, exc, tb):
        def webhook_exit(proc):
            print(f"webhook_server {proc} terminated")

        print("\nstopping webhook_server processes...")
        self.webhook.stop(wait=True, callback=webhook_exit)


@pytest.fixture(scope="module")
def module_webhook_process():
    with WebhookServerProcess() as webhook:
        yield webhook


@pytest.fixture
def webhook_process():
    with WebhookServerProcess() as webhook:
        yield webhook


@pytest.fixture
def webhook_process_class():
    return WebhookServerProcess


@pytest.fixture
def dump():
    def _dump(obj):
        info(json.dumps(obj, indent=2))

    return _dump


@pytest.fixture
def webhook_tunnel_url(webhook):
    url = webhook.tunnel_url()
    assert isinstance(url, str)
    assert url.startswith("http://")
    return url



# moralis streams client CLI

import json
import logging
import sys

import click

from .api import MoralisStreamsApi
from .defaults import REGION_CHOICES, STATUS_CHOICES, STREAMS_URL
from .exception_handler import ExceptionHandler
from .version import __timestamp__, __version__
from .webhook_cli import webhook

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"

logger = logging.getLogger(__name__)
info = logger.info
debug = logger.debug


@click.group(name="msc")
@click.version_option(message=header)
@click.option("-d", "--debug", is_flag=True, help="debug mode")
@click.option(
    "-u",
    "--url",
    type=str,
    envvar="MORALIS_STREAMS_URL",
    show_envvar=True,
    default=STREAMS_URL,
    help="moralis streams API base URL",
)
@click.option(
    "-k",
    "--key",
    type=str,
    envvar="MORALIS_API_KEY",
    show_envvar=True,
    default=None,
    required=True,
    help="API Key",
)
@click.option(
    "-r",
    "--row-limit",
    type=int,
    default=100,
    envvar="MORALIS_STREAMS_API_ROW_LIMIT",
    show_envvar=True,
    show_default=True,
    help="number of items in each result page",
)
@click.option(
    "-p",
    "--page-limit",
    type=int,
    default=10000,
    envvar="MORALIS_STREAMS_API_PAGE_LIMIT",
    show_envvar=True,
    show_default=True,
    help="number of pages allowed in results",
)
@click.option("-v", "--verbose", is_flag=True, help="output more detail")
@click.pass_context
def cli(ctx, url, key, debug, verbose, row_limit, page_limit):
    """Moralis Streams API CLI"""
    if debug:
        level = logging.DEBUG
        if not verbose:
            logging.getLogger("urllib3.connectionpool").setLevel(
                logging.WARNING
            )

    elif verbose:
        level = logging.INFO
    else:
        level = logging.WARNING
    logging.basicConfig(
        level=level, handlers=[logging.StreamHandler(sys.stderr)]
    )

    info(header)

    ctx.obj = dict(ehandler=ExceptionHandler(debug))
    ctx.obj["debug"] = debug
    ctx.obj["verbose"] = verbose
    ctx.obj["api"] = MoralisStreamsApi(
        api_key=key,
        url=url,
        debug=debug,
        row_limit=row_limit,
        page_limit=page_limit,
    )


def output(result):
    if result is None:
        result = {}
    elif isinstance(result, dict):
        pass
    elif isinstance(result, list):
        pass
    elif hasattr(result, "to_dict"):
        result = result.to_dict()
    else:
        raise TypeError(f"unexpected result type {type(result)}: {result}")
    click.echo(json.dumps(result, indent=2))


@cli.command
@click.pass_context
def get_stats(ctx):
    """output beta statistics"""
    output(ctx.obj["api"].get_stats())


@cli.command
@click.option(
    "-x/-X",
    "--exclude-payload/--no-exclude-payload",
    is_flag=True,
    help="exclude payload in response",
)
@click.pass_context
def get_history(ctx, exclude_payload):
    """output event history"""
    output(ctx.obj["api"].get_history(exclude_payload=exclude_payload))


@cli.command
@click.argument("event-id", type=str)
@click.pass_context
def replay_history(ctx, event_id):
    """request resend of history event"""
    output(ctx.obj["api"].replay_history(event_id))


@cli.command
@click.pass_context
def get_settings(ctx):
    """get settings"""
    output(ctx.obj["api"].get_settings())


@cli.command
@click.pass_context
@click.argument("region", type=click.Choice(REGION_CHOICES))
def set_settings(ctx, region):
    """set settings"""
    output(ctx.obj["api"].set_settings(region))


@cli.command
@click.option(
    "-c",
    "--chain-id",
    type=str,
    multiple=True,
    help='The ids of the chains for this stream in hex Ex: ["0x1","0x38"]',
)
@click.option(
    "-t",
    "--topic",
    multiple=True,
    help="The topic0 of the event in hex (required for CONTRACT",
)
@click.option(
    "-A",
    "--all-addresses",
    is_flag=True,
    help="Include events for all addresses (only applied when abi and topic0 is provided)",
)
@click.option(
    "-n",
    "--include-native-txs",
    is_flag=True,
    help="Include native transactions defaults to false (contract only)",
)
@click.option(
    "-l",
    "--include-contract-logs",
    is_flag=True,
    help="Include events for all addresses (only applied when abi and topic0 is provided)",
)
@click.option(
    "-i",
    "--include-internal-txs",
    is_flag=True,
    help="Include events for all addresses (only applied when abi and topic0 is provided)",
)
@click.option(
    "-a", "--abi", type=str, help="The contract abi as a JSON string"
)
@click.option(
    "-o",
    "--option",
    multiple=True,
    type=str,
    help="json string of Advanced Options {topic0: str, filter: str, include_native_txs: bool}",
)
@click.argument("webhook_url", type=str)
@click.argument("description", type=str)
@click.argument("tag", type=str)
@click.pass_context
def create_stream(
    ctx,
    topic,
    all_addresses,
    include_native_txs,
    include_contract_logs,
    include_internal_txs,
    abi,
    option,
    chain_id,
    webhook_url,
    description,
    tag,
):
    """create new stream"""

    api = ctx.obj["api"]
    ret = api.create_stream(
        webhook_url=webhook_url,
        description=description,
        tag=tag,
        topic0=list(topic),
        all_addresses=all_addresses,
        include_native_txs=include_native_txs,
        include_contract_logs=include_contract_logs,
        include_internal_txs=include_internal_txs,
        abi=json.loads(abi),
        advanced_options=list(option),
        chain_ids=list(chain_id),
    )
    output(ret)


@cli.command
@click.argument("stream-id", type=str)
@click.option(
    "-a",
    "--address",
    multiple=True,
    type=str,
    help="address to add (multiples ok)",
)
@click.pass_context
def add_address_to_stream(ctx, stream_id, address):
    """add one or more addresses to the stream identified by stream-id"""
    api = ctx.obj["api"]
    ret = api.add_address_to_stream(stream_id, list(address))
    output(ret)


@cli.command
@click.argument("stream-id", type=str)
@click.option("-a", "--address", type=str, help="address to remove")
@click.pass_context
def delete_address_from_stream(ctx, stream_id, address):
    """delete an address from the stream identified by stream-id"""
    api = ctx.obj["api"]
    ret = api.delete_addresses_from_stream(stream_id, address)
    output(ret)


@cli.command
@click.argument("stream-id", type=str)
@click.pass_context
def delete_stream(ctx, stream_id):
    """delete stream identified by stream-id"""
    api = ctx.obj["api"]
    ret = api.delete_stream(stream_id)
    output(ret)


@cli.command
@click.argument("stream-id", type=str)
@click.pass_context
def get_addresses(ctx, stream_id):
    """list addresses associated with the stream identified by stream-id"""
    api = ctx.obj["api"]
    ret = api.get_addresses(stream_id)
    output(ret)


@cli.command
@click.option(
    "-i", "--stream-id", type=str, help="stream_id, default is all streams"
)
@click.pass_context
def get_streams(ctx, stream_id):
    """list one or all streams"""
    api = ctx.obj["api"]
    if stream_id is None:
        ret = api.get_streams()
    else:
        ret = api.get_stream(stream_id)
    output(ret)


@cli.command
@click.option(
    "-c",
    "--chain-id",
    type=str,
    multiple=True,
    help='The ids of the chains for this stream in hex Ex: ["0x1","0x38"]',
)
@click.option(
    "-t",
    "--topic",
    multiple=True,
    help="The topic0 of the event in hex (required for CONTRACT",
)
@click.option(
    "-A",
    "--all-addresses",
    is_flag=True,
    help="Include events for all addresses (only applied when abi and topic0 is provided)",
)
@click.option(
    "-n",
    "--include-native-txs",
    is_flag=True,
    help="Include native transactions defaults to false (contract only)",
)
@click.option(
    "-l",
    "--include-contract-logs",
    is_flag=True,
    help="Include events for all addresses (only applied when abi and topic0 is provided)",
)
@click.option(
    "-i",
    "--include-internal-txs",
    is_flag=True,
    help="Include events for all addresses (only applied when abi and topic0 is provided)",
)
@click.option(
    "-a",
    "--abi",
    type=str,
    multiple=True,
    help="The abi to parse the log object of the contract",
)
@click.option(
    "-o",
    "--option",
    multiple=True,
    type=str,
    help="json string of Advanced Options {topic0: str, filter: str, include_native_txs: bool}",
)
@click.argument("stream-id", type=str)
@click.argument("webhook-url", type=str)
@click.argument("description", type=str)
@click.argument("tag", type=str)
@click.pass_context
def update_stream(
    ctx,
    stream_id,
    webhook_url,
    description,
    tag,
    topic,
    all_addresses,
    include_native_txs,
    include_contract_logs,
    include_internal_txs,
    abi,
    option,
    chain_id,
):
    """update stream parameters"""
    api = ctx.obj["api"]

    ret = api.update_stream(
        id=stream_id,
        webhook_url=webhook_url,
        description=description,
        tag=tag,
        topic0=list(topic),
        all_addresses=all_addresses,
        include_native_txs=include_native_txs,
        include_contract_logs=include_contract_logs,
        include_internal_txs=include_internal_txs,
        abi=list(abi),
        advanced_options=list(option),
        chain_ids=list(chain_id),
    )
    output(ret)


@cli.command
@click.argument("stream-id", type=str)
@click.argument("status", type=click.Choice(STATUS_CHOICES))
@click.pass_context
def update_stream_status(ctx, stream_id, status):
    """update the status of a stream to active, paused, or error"""
    api = ctx.obj["api"]
    ret = api.update_stream_status(stream_id, status)
    output(ret)


cli.add_command(webhook)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import logging
from typing import Optional, Type

from fastapi.responses import JSONResponse


class ContentSizeLimitExceeded(Exception):
    pass


logger = logging.getLogger(__name__)
debug = logger.debug
critical = logger.debug


class ContentSizeLimitMiddleware:
    """Content size limiting middleware for ASGI applications

    Args:
      app (ASGI application): ASGI application
      max_content_size (optional): the maximum content size allowed in bytes, None for no limit
      exception_cls (optional): the class of exception to raise (ContentSizeExceeded is the default)
    """

    def __init__(
        self,
        app,
        max_content_size: Optional[int] = None,
    ):
        self.app = app
        self.max_content_size = max_content_size

    def receive_wrapper(self, receive):
        received = 0

        async def inner():
            nonlocal received
            message = await receive()
            if (
                message["type"] != "http.request"
                or self.max_content_size is None
            ):
                return message
            body_len = len(message.get("body", b""))
            received += body_len
            if received > self.max_content_size:
                msg = f"Maximum content size limit ({self.max_content_size}) exceeded ({received} bytes read)"
                critical(msg)
                return ContentSizeLimitExceeded(msg)
            return message

        return inner

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        wrapper = self.receive_wrapper(receive)
        await self.app(scope, wrapper, send)



# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from typing import List

from pydantic import BaseModel, Extra

from .model import (
    AbiItem,
    Block,
    InternalTransaction,
    Log,
    Transaction,
    VBaseModel,
)



# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Extra, Field

from .model import SettingsRegion, VBaseModel


class SettingsModel(VBaseModel):

    region: Optional[SettingsRegion] = Field(
        None,
        description="The region from where all the webhooks will be posted for this project",
    )



# ngrok tunnel

import atexit
import logging
import os

from pyngrok import ngrok


class NgrokTunnel:

    tunnel = None

    def __init__(self, port=None, token=None):
        if self.__class__.tunnel is None:
            logging.debug("Connecting ngrok tunnel...")
            token = token or os.environ["NGROK_AUTHTOKEN"]
            ngrok.set_auth_token(token)
            self.__class__.tunnel = ngrok.connect(port)
            logging.debug(f"public_url={self.url()}")

    def __enter__(self):
        return self.tunnel

    def __exit__(self, _, exc, tb):
        if self.tunnel is not None:
            self.disconnect()

    def url(self):
        return self.tunnel.public_url

    def disconnect(self):
        url = self.url()
        logging.debug(f"Disconnecting ngrok tunnel: {url}")
        ngrok.disconnect(url)
        self.__class__.tunnel = None



# api tests

import logging
import os
import sys
import time
from copy import deepcopy
from decimal import Decimal
from pathlib import Path

import MoralisSDK.api
import pytest
import starlette.config
from backoff import expo, on_exception
from box import Box, BoxList
from eth_account import Account
from eth_utils import (
    event_abi_to_log_topic,
    from_wei,
    is_same_address,
    to_checksum_address,
    to_hex,
    to_wei,
)
from ratelimit import RateLimitException, limits

from moralis_streams_client import MoralisStreamsApi, models

# TODO: generate history and multiple streams to ensure enough response data

debug = logging.debug
info = logging.info
logging.getLogger("moralis_streams_client.api").setLevel("WARNING")
logging.getLogger("urllib3.connectionpool").setLevel("WARNING")


ETHERSCAN_LIMIT_COUNT = 5
ETHERSCAN_LIMIT_SECONDS = 1

CONFIRM_COUNT = 12

EVENT_NAME = "PrintMintPending"

CALLBACK_TIMEOUT = 300

CALLBACK_TEST_KEYS = [
    "abi",
    "block",
    "chainId",
    "confirmed",
    "erc20Approvals",
    "erc20Transfers",
    "logs",
    "nftApprovals",
    "nftTransfers",
    "retries",
    "txs",
    "txsInternal",
    "streamId",
    "tag",
]


@pytest.fixture
def event_name():
    return EVENT_NAME


@pytest.fixture
def background_contract_address(config):
    address = config("TEST_BACKGROUND_CONTRACT")
    return to_checksum_address(address)


# rate-limit etherscan api calls
@on_exception(expo, RateLimitException, max_tries=16)
@limits(calls=ETHERSCAN_LIMIT_COUNT, period=ETHERSCAN_LIMIT_SECONDS)
def limit(func, *args, **kwargs):
    return func(*args, **kwargs)


@pytest.fixture
def get_contract_tokens(moralis, network):
    def _get_contract_tokens(contract_address):
        owners = Box(moralis.get_nft_owners(contract_address, chain=network))
        assert "result" in owners
        return owners.result

    return _get_contract_tokens


def test_api_transactions(ape, get_contract_tokens, system_address, moralis):
    assert ape.explorer
    txns = list(limit(ape.explorer.get_account_transactions, system_address))
    assert txns
    contracts = set()
    for txn in txns:
        info(f"{txn.transaction}")
        to_addr = txn.transaction.dict()["to"]
        if to_addr == "":
            assert txn.contract_address
            contracts.add(to_checksum_address(txn.contract_address))

    info(f"txn_count={len(txns)}")
    info(f"contracts={len(contracts)}")
    for addr in contracts:
        contract = ape.Contract(addr)
        info(f"{contract.symbol()} {contract}")

        url = ape.explorer.get_address_url(addr)
        info(url)

        token_ids = Box(moralis.get_token_id(addr, chain="goerli"))
        info(f"  tokens={int(token_ids.total)}")

        tokens = get_contract_tokens(addr)
        assert token_ids.total == len(tokens)
        for token in tokens:
            info(f"  {int(token.token_id)}: {token.name} {token.owner_of}")


@pytest.fixture
def ethersieve_contract(ape, ethersieve_contract_address):
    return ape.Contract(ethersieve_contract_address)


@pytest.fixture
def background_contract(ape, background_contract_address):
    return ape.Contract(background_contract_address)


@pytest.mark.uses_gas
def test_api_create_stream(
    streams,
    dump,
    chain_id,
    ethersieve_contract,
    background_contract,
    event_name,
    event_keys,
    get_contract_tokens,
    system_address,
    system_key,
    user_address,
    user_key,
    ape,
    webhook,
    webhook_tunnel_url,
):
    webhook.clear()

    streams_begin = streams.get_streams()

    event_abi = ethersieve_contract.contract_type.events[event_name].dict()
    event_topic = to_hex(event_abi_to_log_topic(event_abi))
    contract_abi = ethersieve_contract.contract_type.dict()["abi"]

    stream = Box(
        streams.create_stream(
            webhook_url=webhook_tunnel_url + "/contract/event",
            description="moralis_streams_client testing stream",
            tag="msc_test",
            topic0=[event_topic],
            all_addresses=False,
            include_native_txs=True,
            include_contract_logs=True,
            include_internal_txs=True,
            abi=contract_abi,
            advanced_options=None,
            chain_ids=[chain_id],
        )
    )
    dump(f"{stream.id=}")
    dump(f"{stream.tag=}")
    dump(f"{stream.status=}")
    dump(f"{stream.statusMessage=}")
    address_added = Box(
        streams.add_address_to_stream(stream.id, [ethersieve_contract.address])
    )
    assert isinstance(address_added, dict)
    dump(address_added)
    assert "streamId" in address_added
    assert "address" in address_added
    assert isinstance(address_added.streamId, str)
    assert isinstance(address_added.address, list)
    addresses = address_added.address
    for address in addresses:
        assert isinstance(address, str)
        assert is_same_address(address, ethersieve_contract.address)

    background_ids = []
    for token in get_contract_tokens(background_contract.address):
        if is_same_address(token.owner_of, user_address):
            background_ids.append(token.token_id)
    dump(f"user background tokens: {background_ids}")

    ethersieve_ids = []
    for token in get_contract_tokens(ethersieve_contract.address):
        if is_same_address(token.owner_of, user_address):
            ethersieve_ids.append(token.token_id)
    dump(f"user ethersieve tokens: {ethersieve_ids}")

    # send a printMint transaction from user to ethersieve

    PRIORITY_FEE_GWEI = 3

    user = Account().from_key(user_key)
    max_priority_fee = ape.provider.priority_fee + to_wei(
        PRIORITY_FEE_GWEI, "gwei"
    )

    info(f"max_priority_fee={from_wei(max_priority_fee, 'gwei')} gwei")

    args = (0, 0, 1, 2, 3)
    kwargs = dict(
        value=to_wei(Decimal(".01"), "ether"),
        sender=user.address,
        nonce=ape.provider.get_nonce(user.address),
        max_priority_fee=max_priority_fee,
        required_confirmations=0,
    )
    txn = ethersieve_contract.printMint.as_transaction(*args, **kwargs)
    txn.signature = user.sign_transaction(txn.dict())
    estimated_gas = ape.provider.estimate_gas_cost(txn)
    info(f"{estimated_gas=}")
    receipt = ape.provider.send_transaction(txn)

    dump(dict(txn_hash=receipt.txn_hash, status=str(receipt.status)))

    CONFIRM_COUNT = 2

    confirmed_height = receipt.block_number + CONFIRM_COUNT
    last_height = receipt.block_number

    info(f"Waiting for {CONFIRM_COUNT} confirmations...")
    count = 1
    while last_height < confirmed_height:
        height = ape.provider.chain_manager.blocks.height
        if last_height != height:
            info(f"block[{count} of {CONFIRM_COUNT}]: {height}")
            last_height = height
            count += 1

    info(f"Waiting up to {CALLBACK_TIMEOUT} seconds for 3 callback events...")
    events = {}

    timeout = time.time() + CALLBACK_TIMEOUT
    while len(events) < 3:
        assert time.time() < timeout, "timeout waiting for 3 callback events"
        all_events = webhook.events()
        for event in all_events:
            event_id = event["id"]
            if event_id not in events:
                events[event_id] = event
                info(f"{event_id=}")
        time.sleep(1)

    events = webhook.events()
    assert isinstance(events, list)
    for event in events:
        assert isinstance(event, dict)
        assert set(event.keys()) == set(event_keys)

    assert set(events[0]["body"].keys()) == set(CALLBACK_TEST_KEYS)
    unconfirmed = events[1]["body"]
    assert unconfirmed["confirmed"] is False
    confirmed = events[2]["body"]
    assert confirmed["confirmed"] is True
    for e in [unconfirmed, confirmed]:
        assert len(e["txs"]) == 1
        assert is_same_address(e["txs"][0]["fromAddress"], user_address)
        assert is_same_address(
            e["txs"][0]["toAddress"], ethersieve_contract.address
        )

    result = Box(streams.delete_stream(stream.id))
    assert result.id == stream.id

    streams_end = streams.get_streams()

    assert len(streams_begin) == len(streams_end)


def test_api_get_streams_default(streams):
    all_streams = streams.get_streams()
    assert isinstance(all_streams, list)
    for i, stream in enumerate(all_streams):
        assert isinstance(stream, dict)
        info(f"{i}: {stream['id']}")


def test_api_get_streams_row_limit(streams, monkeypatch):
    # set low row limit to ensure paging
    monkeypatch.setattr(streams, "row_limit", 3)
    streams = streams.get_streams()
    assert isinstance(streams, list)
    for i, stream in enumerate(streams):
        assert isinstance(stream, dict)
        info(f"{i}: {stream['id']}")


def test_api_get_streams_row_page_limit(streams, monkeypatch):
    # set low row and page limit to ensure paging
    monkeypatch.setattr(streams, "row_limit", 2)
    monkeypatch.setattr(streams, "page_limit", 2)
    all_streams = streams.get_streams()
    assert isinstance(all_streams, list)
    for i, stream in enumerate(all_streams):
        assert isinstance(stream, dict)
        info(f"{i}: {stream['id']}")


def test_api_get_history_default(streams):
    history_events = streams.get_history()
    assert isinstance(history_events, list)
    for i, history_event in enumerate(history_events):
        assert isinstance(history_event, dict)
        info(f"{i}: {history_event['id']}")


def test_api_get_history_row_limit(streams, monkeypatch):
    # set low row limit to ensure paging
    monkeypatch.setattr(streams, "row_limit", 3)
    history_events = streams.get_history()
    assert isinstance(history_events, list)
    for i, history_event in enumerate(history_events):
        assert isinstance(history_event, dict)
        info(f"{i}: {history_event['id']}")


def test_api_get_history_row_page_limit(streams, monkeypatch):
    # set low row and page limit to ensure paging
    monkeypatch.setattr(streams, "row_limit", 3)
    monkeypatch.setattr(streams, "page_limit", 2)
    history_events = streams.get_history()
    assert isinstance(history_events, list)
    for i, history_event in enumerate(history_events):
        assert isinstance(history_event, dict)
        info(f"{i}: {history_event['id']}")


def test_api_delete_all_streams(streams, dump):
    all_streams = streams.get_streams()
    assert isinstance(all_streams, list)
    for stream in all_streams:
        assert isinstance(stream, dict)
        stream_id = stream["id"]
        dump(f"deleting stream {stream_id}")
        result = streams.delete_stream(stream_id)
        assert isinstance(result, dict)
        assert result["id"] == stream_id

    all_streams = streams.get_streams()
    assert isinstance(all_streams, list)
    assert len(all_streams) == 0



# streams api client wrapper

import json
import logging
from pprint import pformat

from fastapi import HTTPException, Request
from requests import codes
from starlette.middleware.base import BaseHTTPMiddleware

from . import settings

logger = logging.getLogger("signature")
debug = logger.debug
error = logger.error
critical = logger.critical


async def validate_signature(request: Request):

    signature = request.app.state.signature
    path = request.scope["path"]

    if signature is None or signature.header is None or signature.key is None:
        msg = "signature validator not configured"
        critical(f"{msg}: {path}")
        raise HTTPException(
            detail=msg, status_code=codes.INTERNAL_SERVER_ERROR
        )

    request_signature = request.headers.get(signature.header, None)

    if request_signature is None:
        msg = "missing signature"
        error(f"{msg}: {path}")
        raise HTTPException(detail=msg, status_code=codes.BAD_REQUEST)
    else:
        body = await request.body()
        if len(body) == 0:
            body = b""
        else:
            body = json.dumps(json.loads(body), separators=(",", ":")).encode()
        debug(
            f"validating: {path} type={type(body)} length={len(body)} {request_signature}"
        )
        if not signature.validate(request_signature, body):
            msg = "authorization failed"
            error(f"{msg}: {path}")
            raise HTTPException(detail=msg, status_code=codes.UNAUTHORIZED)



# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel, Extra, Field

from .model import (
    AbiItem,
    AdvancedOptions,
    StreamsModel,
    StreamsStatus,
    VBaseModel,
)


class StreamsResponse(VBaseModel):

    result: List[StreamsModel] = Field(
        ..., description="Array of project Streams"
    )
    cursor: Optional[str] = Field(
        None, description="Cursor for fetching next page"
    )
    total: float = Field(
        ..., description="Total count of streams on the project"
    )


class StreamsModelCreate(VBaseModel):

    webhookUrl: str = Field(
        ...,
        description="Webhook URL where moralis will send the POST request.",
    )
    description: str = Field(..., description="A description for this stream")
    tag: str = Field(
        ...,
        description="A user-provided tag that will be send along the webhook, the user can use this tag to identify the specific stream if multiple streams are present",
    )
    topic0: Optional[List[str]] = Field(
        None,
        description="An Array of topic0's in hex, required if the type : log",
    )
    allAddresses: Optional[bool] = Field(
        None,
        description="Include events for all addresses (only applied when abi and topic0 is provided)",
    )
    includeNativeTxs: Optional[bool] = Field(
        None,
        description="Include or not native transactions defaults to false (only applied when type:contract)",
    )
    includeContractLogs: Optional[bool] = Field(
        None,
        description="Include or not logs of contract interactions defaults to false",
    )
    includeInternalTxs: Optional[bool] = Field(
        None,
        description="Include or not include internal transactions defaults to false",
    )
    abi: Optional[List[AbiItem]] = None
    advancedOptions: Optional[List[AdvancedOptions]] = None
    chainIds: List[str] = Field(
        ...,
        description='The ids of the chains for this stream in hex Ex: ["0x1","0x38"]',
    )


class StreamsStatusUpdate(VBaseModel):

    status: StreamsStatus = Field(..., description="The status of the stream.")



# streams api client wrapper

import json
import logging
import os
from pprint import pformat
from typing import Dict, List

import requests
from requests.exceptions import HTTPError, JSONDecodeError

from . import settings
from .defaults import (
    ACTIVE,
    ERROR,
    PAGE_LIMIT,
    PAUSED,
    REGION,
    REGION_CHOICES,
    ROW_LIMIT,
    STREAMS_URL,
)
from .exceptions import (
    MoralisStreamsCallFailed,
    MoralisStreamsErrorReturned,
    MoralisStreamsResponseFormatError,
)

logger = logging.getLogger(__name__)
info = logger.info
debug = logger.debug
error = logging.error


class MoralisStreamsApi:
    def __init__(
        self,
        *,
        api_key=None,
        url=STREAMS_URL,
        region=REGION,
        debug=None,
        row_limit=ROW_LIMIT,
        page_limit=PAGE_LIMIT,
    ):
        self.api_key = str(settings.MORALIS_API_KEY)
        self.url = url
        self.headers = {"x-api-key": api_key}
        self.page_limit = page_limit
        self.row_limit = row_limit
        self.debug = debug or settings.MORALIS_STREAMS_API_DEBUG
        self._init_region(region)

    def _env_flag(self, name, default=False):
        default = "1" if default else "0"
        return bool(int(os.environ.get(name, default)))

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"{self.__class__.__name__}<{hex(id(self))}>"

    def _init_region(self, region):
        self.get_settings()
        if self.region != region:
            self.set_settings(region=region)

    def _parse_advanced_options(self, advanced_options):
        options = advanced_options or []
        for o in options:
            odict = json.loads(o)
            options.append(
                {
                    key: odict[key]
                    for key in ["topic0", "filter", "includeNativeTxs"]
                }
            )
        return options

    def _get(self, path, *, params={}, paginated=False, require_keys=[]):

        if paginated:
            return self._get_paginated(path, params)

        response = requests.get(
            self.url + path, headers=self.headers, params=params
        )
        return self._return_result(response, require_keys)

    def _post(self, path, body={}):
        response = requests.post(
            self.url + path, headers=self.headers, json=body
        )
        return self._return_result(response)

    def _put(self, path, body={}):
        response = requests.put(
            self.url + path, headers=self.headers, json=body
        )
        return self._return_result(response)

    def _delete(self, path, body={}):
        response = requests.delete(
            self.url + path, headers=self.headers, json=body
        )
        return self._return_result(response)

    def kludge(self, message):
        """patch buggy return value seen on 2022-10-11 in POST streams/evm/{streamID}"""
        error(f"kludge: {message=}")
        if "0" in message and "1" in message and message["0"] == "id":
            if self.debug:
                error('patching "0":"id", "1":"<stream_id>"')
            message[message.pop("0")] == message.pop("1")
        return message

    def _return_result(self, response, require_keys=[]):

        logger.debug(f"_return_result {self}")

        errors = []
        try:
            message = response.json()
        except JSONDecodeError as exc:
            message = response.text
            errors.append(MoralisStreamsResponseFormatError(f"{exc}"))

        if settings.MORALIS_STREAMS_API_PROTOCOL_PATCH:
            message = self.kludge(message)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            errors.append(str(exc))
            raise MoralisStreamsCallFailed((message, errors)) from exc

        if isinstance(message, dict):
            for key in require_keys:
                if key not in message:
                    errors.append(f"missing required key '{key}'")

        if errors:
            raise MoralisStreamsResponseFormatError((message, errors))

        if response.ok is False:
            raise MoralisStreamsErrorReturned((message, response.reason))

        return message

    def _get_page(self, count, path, params, results, require_keys):

        if self.debug:
            debug("----")
            debug(f"{count=}")
            debug(f"get({repr(path)},{params=})")

        ret = self._get(path, params=params, require_keys=require_keys)

        if self.debug:
            debug(f"  total={ret['total']}")
            debug(f"  cursor={ret.get('cursor', '<NOT_PRESENT>')}")

        result = ret.get("result")

        if self.debug:
            debug(f"  result=[{len(result)}]")
            for i, r in enumerate(result):
                debug(f"    result[{i}]: {r}")

        if len(result) > 0:
            if self.debug:
                len_before = len(results)

            results.extend(result)

            if self.debug:
                len_after = len(results)
                debug(f"results: [{len_before}] -> [{len_after}]")
                for i, r in enumerate(results):
                    debug(f"  results[{i}]: {r}")

        return ret, results

    def _get_paginated(self, path, params={}):
        results = []
        params.setdefault("limit", self.row_limit)
        count = 0
        cursor = None
        total = None
        if self.debug:
            debug("---BEGIN_PAGINATED---")
        while True:
            if cursor:
                params["cursor"] = cursor
            else:
                params.pop("cursor", None)

            ret, results = self._get_page(
                count, path, params, results, ["total", "result"]
            )

            _total = int(ret["total"])
            if total is None:
                total = _total
                if self.debug:
                    debug(f"setting {total=} on iteration {count=}")
            else:
                if total != _total:
                    raise MoralisStreamsResponseFormatError(
                        f"total_changed: original={total=} latest={_total} {count=}"
                    )

            if len(results) == total:
                if self.debug:
                    debug(f"pagination_exit: {total=} results={len(results)}")
                break

            # missing cursor exit
            try:
                cursor = ret["cursor"]
            except KeyError:
                if self.debug:
                    debug("pagination_exit: no cursor in ret")
                break

            # null or empty string cursor exit
            if cursor in ["", None]:
                error(f"NULL cursor returned: NULL cursor={repr(cursor)}")

            # check for total overrun
            if len(results) > total:
                raise MoralisStreamsResponseFormatError(
                    f"overrun: {total=} results_len={len(results)}"
                )

            # check for runaway page count
            count += 1
            if count > self.page_limit:
                raise MoralisStreamsResponseFormatError(
                    f"exceeded page count limit ({self.page_limit})"
                )

        if total is None:
            raise MoralisStreamsResponseFormatError("exited with {total=}")

        if len(results) != total:
            raise MoralisStreamsResponseFormatError(
                f"results length mismatch: {total=} len(results)={len(results)}"
            )

        if self.debug:
            debug("---END_PAGINATED---")

        return results

    def get_stats(self) -> dict:
        debug(f"{self} get_stats()")
        ret = self._get("/beta/stats")
        debug(f"{self} {ret=}")
        return ret

    def get_settings(self) -> dict:
        debug(f"{self} get_settings()")
        settings = self._get("/settings", require_keys=["region"])
        self.region = settings["region"]
        ret = None
        debug(f"{self} {ret=}")
        return ret

    def set_settings(self, region: str) -> None:
        debug(f"{self} set_settings({region=})")
        self._post("/settings", dict(region=region))
        self.region = region
        ret = None
        debug(f"{self} {ret=}")
        return ret

    def create_stream(
        self,
        *,
        webhook_url: str,
        description: str,
        tag: str,
        topic0: List[str],
        all_addresses: bool,
        include_native_txs: bool,
        include_contract_logs: bool,
        include_internal_txs: bool,
        abi: List[Dict],
        advanced_options: List[Dict],
        chain_ids: List[str],
    ) -> dict:

        params = dict(
            webhookUrl=webhook_url,
            description=description,
            tag=tag,
            topic0=topic0,
            allAddresses=all_addresses,
            includeNativeTxs=include_native_txs,
            includeContractLogs=include_contract_logs,
            includeInternalTxs=include_internal_txs,
            abi=abi,
            advancedOptions=self._parse_advanced_options(advanced_options),
            chainIds=chain_ids,
        )
        debug(f"{self} create_stream({params=})")
        ret = self._put("/streams/evm", params)
        debug(f"{self} {ret=}")
        return ret

    def add_address_to_stream(self, stream_id: str, address: str) -> dict:
        debug(f"{self} add_address_to_stream({stream_id=}, {address=})")
        path = f"/streams/evm/{stream_id}/address"
        params = dict(address=address)
        ret = self._post(path, params)
        debug(f"{self} {ret=}")
        return ret

    def delete_address_from_stream(self, stream_id: str, address: str) -> dict:
        debug(f"{self} delete_address_from_stream({stream_id=}, {address=})")
        path = f"/streams/evm/{stream_id}/address"
        params = dict(address=address)
        ret = self._delete(path, params)
        debug(f"{self} {ret=}")
        return ret

    def delete_stream(self, stream_id: str) -> dict:
        debug(f"{self} delete_stream({stream_id=})")
        path = f"/streams/evm/{stream_id}"
        ret = self._delete(path)
        debug(f"{self} {ret=}")
        return ret

    def get_addresses(self, stream_id: str) -> List[str]:
        debug(f"{self} get_addresses({stream_id=})")
        path = f"/streams/evm/{stream_id}/address"
        ret = self._get(path, paginated=True)
        debug(f"{self} {ret=}")
        return ret

    def get_stream(self, stream_id: str) -> dict:
        debug(f"{self} get_stream({stream_id=})")
        path = f"/streams/evm/{stream_id}"
        ret = self._get(path)
        debug(f"{self} {ret=}")
        return ret

    def get_streams(self) -> List[Dict]:
        debug(f"{self} get_streams()")
        path = "/streams/evm"
        ret = self._get(path, paginated=True)
        debug(f"{self} {ret=}")
        return ret

    def update_stream(
        self,
        stream_id: str,
        *,
        webhook_url: str = None,
        description: str = None,
        tag: str = None,
        topic0: List[str] = None,
        all_addresses: bool = None,
        include_native_txs: bool = None,
        include_contract_logs: bool = None,
        include_internal_txs: bool = None,
        abi: List[str] = None,
        advanced_options: List[str] = None,
        chain_ids: List[str] = None,
    ) -> dict:
        path = f"/streams/evm/{stream_id}"

        params = dict(
            webhookUrl=webhook_url,
            description=description,
            tag=tag,
            topic0=topic0,
            allAddresses=all_addresses,
            includeNativeTxs=include_native_txs,
            includeContractLogs=include_contract_logs,
            includeInternalTxs=include_internal_txs,
            abi=abi,
            advancedOptions=self._parse_advanced_options(advanced_options),
            chainIds=chain_ids,
        )
        debug(f"{self} update_stream({stream_id=}, {params=})")
        ret = self._post(path, params)
        debug(f"{self} {ret=}")
        return ret

    def update_stream_status(self, stream_id: str, status: str) -> dict:
        debug(f"{self} update_stream_status({stream_id=}, {status=})")
        path = f"/streams/evm/{stream_id}/status"
        params = dict(status=status)
        ret = self._post(path, params)
        debug(f"{self} {ret=}")
        return ret

    def get_history(
        self,
        exclude_payload: bool = False,
    ) -> dict:
        debug(f"{self} get_history({exclude_payload=})")
        path = "/history"
        if exclude_payload is True:
            params = dict(excludePayload=exclude_payload)
        else:
            params = {}
        ret = self._get(path, params=params, paginated=True)
        debug(f"{self} {ret=}")
        return ret

    def replay_history(self, event_id: str) -> List[Dict]:
        debug(f"{self} replay_history({event_id=})")
        path = f"/history/replay/{event_id}"
        ret = self._post(path)
        debug(f"{self} {ret=}")
        return ret



# test relay mode

import atexit
import time
from logging import critical, info
from pathlib import Path
from subprocess import run
from uuid import uuid4

import pytest
from requests.structures import CaseInsensitiveDict as Headers

from moralis_streams_client import settings
from moralis_streams_client.app import obscure_key
from moralis_streams_client.webhook import Webhook

EVENT_COUNT = 10

TEST_RELAY_KEY = "test_relay_key"
RELAY_KEY_HEADER = "X-API-Key"
RELAY_PORT = 8080
TARGET_PORT = 8081


@pytest.fixture(scope="module", autouse=True)
def webhook_process(module_webhook_process):
    yield module_webhook_process


@pytest.fixture
def relay_port():
    return RELAY_PORT


@pytest.fixture
def target_port():
    return TARGET_PORT


@pytest.fixture(scope="module")
def verify_webhook():
    def _verify_webhook(_webhook, port):
        assert _webhook.server_running() is True
        procs = _webhook.processes()
        assert isinstance(procs, list)
        assert len(procs) > 0
        assert _webhook.port == port
        assert _webhook.clear()
        return True

    return _verify_webhook


@pytest.fixture
def relay_url(target_port):
    return f"http://localhost:{target_port}/contract/event"


@pytest.fixture
def relay_key():
    return TEST_RELAY_KEY


@pytest.fixture
def relay_header():
    return RELAY_KEY_HEADER


@pytest.fixture(scope="module")
def relay(webhook, verify_webhook):
    """configue running webhook to relay messages to target webhook"""
    relay_port = RELAY_PORT
    target_port = TARGET_PORT
    relay_key = TEST_RELAY_KEY
    relay_header = RELAY_KEY_HEADER
    relay_url = f"http://localhost:{target_port}/contract/event"

    _relay = webhook
    verify_webhook(_relay, relay_port)
    set_result = _relay.relay(
        url=relay_url, key=relay_key, header=relay_header
    )
    assert isinstance(set_result, dict)
    get_result = _relay.relay()
    assert get_result == dict(
        url=relay_url, header=relay_header, key=obscure_key(relay_key)
    )
    try:
        yield _relay
    finally:
        assert _relay.events() == []
        _relay.relay(enable=False)
        assert _relay.relay() == dict(url=None, header=None, key=None)
        assert _relay.buffer() is True


@pytest.fixture(scope="module")
def target(wait_for_it, verify_webhook):
    target_port = TARGET_PORT
    _target = Webhook(
        debug=True,
        port=target_port,
        tunnel=False,
    )
    logfile = Path(".") / "relay.log"
    _target.start(wait=True, logfile=str(logfile))
    wait_for_it(_target.addr, _target.port)
    assert _target.server_running()
    assert verify_webhook(_target, target_port)
    _target.relay(enable=False)
    try:
        yield _target
    finally:
        assert _target.events() == []
        _target.relay(enable=False)
        assert _target.relay() == dict(url=None, header=None, key=None)
        assert _target.buffer() is True
        _target.stop()


def test_relay_messages(relay, target, relay_url, relay_key, relay_header):
    assert relay.clear()
    assert target.clear()

    assert relay.relay() == dict(
        url=f"http://localhost:{target.port}/contract/event",
        header=relay_header,
        key=obscure_key(relay_key),
    )
    assert target.relay() == dict(url=None, header=None, key=None)

    assert len(target.events()) == 0
    assert len(relay.events()) == 0

    sent_events = {}
    for i in range(EVENT_COUNT):
        event = dict(test=f"payload_{i}", count=i, id=str(uuid4()))
        event_id = relay.inject(event)
        sent_events[event_id] = event

    while len(target.events()) < EVENT_COUNT:
        info(f"target_events_length={target.events()}")
        time.sleep(1)

    target_events = target.events()

    sent_ids = list(sent_events.keys())
    target_ids = [t["id"] for t in target_events]
    assert len(sent_ids) == len(target_ids)

    for i, sid in enumerate(sent_events.keys()):
        sbody = sent_events[sid]
        tev = target_events[i]
        tbody = tev["body"]
        assert sbody == tbody
        assert isinstance(tev["headers"], dict)
        assert relay_header in Headers(tev["headers"])
        assert Headers(tev["headers"])[relay_header] == relay_key

    for event in target_events:
        event_id = Headers(event["headers"])["x-relay-id"]
        assert relay.event(event_id)
        relay.delete(event_id)

    assert relay.events() == []


def test_relay_id(relay, target, relay_url, relay_key, relay_header, dump):
    assert relay.clear()
    assert target.clear()

    msg_id = str(uuid4())
    event = dict(test="payload", id=msg_id)
    relay.inject(event)
    tevs = target.events()
    assert len(tevs) == 1
    dump(tevs[0])
    assert tevs[0]["body"]["id"] == msg_id
    relay.clear()
    target.clear()


def _make_huge_event(size):
    msg_id = str(uuid4())
    huge = "a" + "u" * size + "gh"
    event = dict(test="huge", id=msg_id, castle_of=huge)
    return event


def test_relay_huge(relay, target, relay_url, relay_key, relay_header, dump):
    assert relay.clear()
    assert target.clear()

    event = _make_huge_event(8_000_000)

    relay.inject(event)

    tevs = target.events()
    assert len(tevs) == 1
    assert tevs[0]["body"]["id"] == event["id"]
    assert len(tevs[0]["body"]["castle_of"]) == len(event["castle_of"])
    relay.clear()
    target.clear()


def test_relay_recover_huge(
    relay, target, relay_url, relay_key, relay_header, dump
):
    assert relay.clear()
    assert target.clear()

    event = _make_huge_event(11_000_000)

    with pytest.raises(Exception) as exc:
        relay.inject(event)
    info(f"Exception: {exc}")

    event = _make_huge_event(9_000_000)
    relay.inject(event)

    tevs = target.events()
    assert len(tevs) == 1
    assert tevs[0]["body"]["id"] == event["id"]
    assert len(tevs[0]["body"]["castle_of"]) == len(event["castle_of"])
    relay.clear()
    target.clear()



# webhook server process

import json
import logging
import os
import re
import sys
import time
from pathlib import Path
from subprocess import DEVNULL, STDOUT, CalledProcessError, Popen, check_output

import psutil
import requests

from .signature import Signature

logger = logging.getLogger(__name__)
debug = logger.debug

PROCESS_START_TIMEOUT = 5
PROCESS_STOP_TIMEOUT = 10


class Webhook:
    def __init__(
        self,
        *,
        addr="127.0.0.1",
        port=8080,
        debug=False,
        base_url=None,
        tunnel=True,
        relay_url=None,
        relay_key=None,
        relay_header=None,
        enable_buffer=True,
        moralis_api_key=None,
    ):
        self.debug = debug
        self.addr = addr
        self.port = port
        base_url = base_url or f"http://{addr}:{port}/"
        self.base_url = base_url.strip("/") + "/"
        self.tunnel = tunnel
        self.relay_url = relay_url
        self.relay_key = relay_key
        self.relay_header = relay_header
        self.enable_buffer = enable_buffer
        self.moralis_api_key = moralis_api_key
        self.signature = Signature()
        self.proc = None

    def _request(self, method, path, **kwargs):
        # generate a signature checksum
        raise_for_status = kwargs.pop("raise_for_status", True)
        kwargs.setdefault("json", None)
        kwargs.setdefault("headers", {})
        if kwargs["json"] is None:
            body = b""
        else:
            body = kwargs["json"]
        kwargs["headers"].update(self.signature.headers(body))

        self.response = requests.request(
            method, self.base_url + path, **kwargs
        )
        if raise_for_status:
            self.response.raise_for_status()
        body = self.response.json()
        return body["result"]

    def hello(self):
        """send and receive a friendly greeting"""
        return self._request("GET", "hello")

    def clear(self):
        """clear the event buffer"""
        return self._request("GET", "clear")

    def tunnel_url(self):
        """return the tunnel url"""
        return self._request("GET", "tunnel")

    def buffer(self, enable=None):
        """set buffer enable"""
        if enable is None:
            method = "GET"
            args = {}
        else:
            method = "POST"
            args = dict(enable=enable)

        return self._request(method, "buffer", json=args)

    def relay(self, url=None, key=None, header=None, enable=None):
        """update relay configuraton"""
        args = {}
        method = "GET"
        if enable is None and url is not None:
            enable = True
        if enable is True:
            method = "POST"
            args["url"] = url
            args["key"] = key
            args["header"] = header
        elif enable is False:
            method = "POST"
            args["url"] = None
            args["key"] = None
            args["header"] = None

        return self._request(method, "relay", json=args)

    def inject(self, event):
        """process an event as a received callback"""
        return self._request("POST", "contract/event", json=event)

    def event(self, event_id):
        """return event by id"""
        return self._request("GET", f"event/{event_id}")

    def delete(self, event_id):
        """delete event by id"""
        return self._request("DELETE", f"event/{event_id}")

    def events(self):
        """return a list of all events"""
        return self._request("GET", "events")

    def shutdown(self):
        """request server shutdown"""
        return self._request("GET", "shutdown")

    def start(self, logfile=None, wait=True):
        """start a server process"""
        if logfile is None:
            logfile = DEVNULL
        else:
            logfile = Path(logfile).open("a") or DEVNULL
        cmd = ["msc"]
        cmd.extend(["webhook", "--addr", self.addr, "--port", str(self.port)])
        if self.debug:
            cmd.append("--debug")
        if self.tunnel:
            cmd.append("--tunnel")
        else:
            cmd.append("--no-tunnel")
        env = os.environ.copy()
        if self.enable_buffer:
            cmd.append("--enable-buffer")
        else:
            cmd.append("--disable-buffer")
        if self.relay_url:
            cmd.extend(["--relay-url", self.relay_url])
        if self.relay_header:
            cmd.extend(["--relay-header", self.relay_header])
        if self.relay_key:
            env["WEBHOOK_RELAY_KEY"] = self.relay_key
        if self.moralis_api_key:
            env["WEBHOOK_API_KEY"] = self.moralis_api_key

        env["PYTHONUNBUFFERED"] = "1"

        cmd.append("server")

        debug(f"command: {cmd}")
        for k, v in env.items():
            if k.startswith("WEBHOOK"):
                debug(f"  {k}={v}")

        self.proc = Popen(
            cmd,
            env=env,
            stderr=STDOUT,
            stdout=logfile,
            bufsize=1,
        )

        if wait:
            timeout = time.time() + PROCESS_START_TIMEOUT
            while self.server_running() is False:
                time.sleep(0.25)
                if time.time() > timeout:
                    raise TimeoutError

        return self.proc

    def processes(self):
        """return list of server processess"""
        pattern = f".*webhook.*--port {self.port}.*"
        procs = []
        for p in psutil.process_iter():
            cmdline = " ".join(p.cmdline())
            if re.match(pattern, cmdline):
                procs.append(p)
        return procs

    def server_running(self):
        """return bool indicating if server is running"""
        return bool(self.processes())

    def stop(self, wait=True, callback=None):
        """signal running server processes to terminate"""
        if self.server_running():
            procs = self.processes()
            for p in procs:
                p.terminate()

            if wait:
                gone, alive = psutil.wait_procs(
                    self.processes(),
                    timeout=PROCESS_STOP_TIMEOUT,
                    callback=callback,
                )
                for p in alive:
                    p.kill()

                if self.server_running():
                    self.shutdown()

            timeout = time.time() + PROCESS_STOP_TIMEOUT
            while wait and self.server_running():
                time.sleep(0.25)
                if time.time() > timeout:
                    raise TimeoutError



# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field

from .model import AbiItem, AdvancedOptions, VBaseModel


class StreamsModelUpdate(VBaseModel):
    webhookUrl: Optional[str] = Field(
        None,
        description="Webhook URL where moralis will send the POST request.",
    )
    description: Optional[str] = Field(
        None, description="A description for this stream"
    )
    tag: Optional[str] = Field(
        None,
        description="A user-provided tag that will be send along the webhook, the user can use this tag to identify the specific stream if multiple streams are present",
    )
    topic0: Optional[List[str]] = Field(
        None,
        description="An Array of topic0's in hex, required if the type : log",
    )
    allAddresses: Optional[bool] = Field(
        None,
        description="Include events for all addresses (only applied when abi and topic0 is provided)",
    )
    includeNativeTxs: Optional[bool] = Field(
        None,
        description="Include or not native transactions defaults to false (only applied when type:contract)",
    )
    includeContractLogs: Optional[bool] = Field(
        None,
        description="Include or not logs of contract interactions defaults to false",
    )
    includeInternalTxs: Optional[bool] = Field(
        None,
        description="Include or not include internal transactions defaults to false",
    )
    abi: Optional[List[AbiItem]] = None
    advancedOptions: Optional[List[AdvancedOptions]] = None
    chainIds: Optional[List[str]] = Field(
        None,
        description='The ids of the chains for this stream in hex Ex: ["0x1","0x38"]',
    )



# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2022-10-16T17:59:39+00:00

from __future__ import annotations

from datetime import datetime
from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel, Extra, Field

from .model import HistoryModel, VBaseModel


class HistoryResponse(VBaseModel):

    result: List[HistoryModel]
    cursor: Optional[str] = None
    total: float
